/*
 * g3 output routines for hp2pbm - bob@usixth.sublink.org. Mon Jan 20
 * 20:44:06 EST 1992
 * 
 * From: 1.1 92/03/05
 */
/*
 * pbmtog3.c - read a portable bitmap and produce a Group 3 FAX file *
 * 
 * Copyright (C) 1989 by Paul Haeberli <paul@manray.sgi.com>. *
 * 
 * Permission to use, copy, modify, and distribute this software and its *
 * documentation for any purpose and without fee is hereby granted, provided *
 * that the above copyright notice appear in all copies and that both that *
 * copyright notice and this permission notice appear in supporting *
 * documentation.  This software is provided "as is" without express or *
 * implied warranty.
 */

#include <stdio.h>
#include "g3.h"

#ifdef	__STDC_
#define	ARGS(a)	a
#else
#define	ARGS(a)	()
#endif

static void putwhitespan ARGS((int c));
static void putblackspan ARGS((int c));
static void putcode ARGS((tableentry * te));
static void puteol ARGS((void));
static void putinit ARGS((void));
static void putbit ARGS((int d));
static void flushbits ARGS((void));

static int      reversebits;
static FILE     *fp;
static int      bigcols;

static int      shdata;
static int      shbit;

static
#ifdef	__GNUC__
inline
#endif
void
putbit(d)
	int             d;
{
	if (d)
		shdata = shdata | shbit;
	if (reversebits)
		shbit = shbit << 1;
	else
		shbit = shbit >> 1;
	if ((shbit & 0xff) == 0) {
		putc(shdata,fp);
		shdata = 0;
		shbit = reversebits ? 0x01 : 0x80;
	}
}

g3start(filep, xmax, rev)
	FILE            *filep;
{
	char            bitrow[1728 / 8];

	fp = filep;
	reversebits = rev;
	bigcols = xmax;
	if (bigcols > 1728)
		bigcols = 1728;

	shdata = 0;
	shbit = reversebits ? 0x01 : 0x80;

	memset(bitrow, 0, sizeof(bitrow));
	g3write(bitrow);
	g3write(bitrow);
	g3write(bitrow);
	g3write(bitrow);
}

g3stop()
{
	puteol();
	puteol();
	puteol();
	puteol();
	puteol();
	puteol();
	flushbits();
}

g3write(bitrow)
	register unsigned char *bitrow;
{
	register int             n = bigcols;
	register int             c;
	register unsigned char	mask, byte;

	byte = *bitrow++;
	mask = 0x80;
	while (n > 0) {
		for ( c=0; (byte&mask) == 0 && n > 0; c++, n-- )
		{
			mask >>= 1;
			if ( mask == 0 )
			{
				mask = 0x80;
				byte = *bitrow++;
				while ( n >= 8 && byte == 0 )
				{
					c += 8;
					n -= 8;
					byte = *bitrow++;
				}
			}
		}
		putwhitespan(c);
		if (n == 0)
			break;
		for ( c=0; (byte&mask) && n > 0; c++, n-- )
		{
			mask >>= 1;
			if ( mask == 0 )
			{
				mask = 0x80;
				byte = *bitrow++;
				while ( n >= 8 && byte == 0xff )
				{
					c += 8;
					n -= 8;
					byte = *bitrow++;
				}
			}
		}
		putblackspan(c);
	}
	puteol();
	return 1;
}

static
#ifdef	__GNUC__
inline
#endif
void
putwhitespan(c)
	int             c;
{
	int             tpos;
	tableentry      *te;

	if (c >= 64) {
		tpos = (c / 64) - 1;
		te = mwtable + tpos;
		c -= te->count;
		putcode(te);
	}
	tpos = c;
	te = twtable + tpos;
	putcode(te);
}

static
#ifdef	__GNUC__
inline
#endif
void
putblackspan(c)
	int             c;
{
	int             tpos;
	tableentry      *te;

	if (c >= 64) {
		tpos = (c / 64) - 1;
		te = mbtable + tpos;
		c -= te->count;
		putcode(te);
	}
	tpos = c;
	te = tbtable + tpos;
	putcode(te);
}

static void
putcode(te)
	register tableentry      *te;
{
	register unsigned int    mask;
	register int             code;

	mask = 1 << (te->length - 1);
	code = te->code;
	while (mask) {
		if (code & mask)
			putbit(1);
		else
			putbit(0);
		mask >>= 1;
	}

}

static void
puteol()
{
	int             i;

	for (i = 0; i < 11; ++i)
		putbit(0);
	putbit(1);
}

static void
flushbits()
{
	if ((reversebits && shbit != 0x01) ||
	    (!reversebits && shbit != 0x80)) {
		putc(shdata,fp);
		shdata = 0;
		shbit = reversebits ? 0x01 : 0x80;
	}
}
